@@ -3,3 +3,5 @@ from .cir import CIR
 from .pubchem import PubChem
 from .chemspider import ChemSpider
 from .opsin import Opsin
+from .cas import CAS
+from .stout import STOUT
